from __future__ import print_function, division
import gc
import numpy as np
import pandas as pd
import scipy.stats as st
import ypy.fs as fs

# make idependent blocks
def collapse_to_chunks(ldblocks, numerators, denominators, numblocks):
    # define endpoints of chunks
    ldblocks.M_H.fillna(0, inplace=True)
    totalM = ldblocks.M_H.sum()
    chunksize = totalM / numblocks
    avgldblocksize = totalM / (ldblocks.M_H != 0).sum()
    chunkendpoints = [0]
    currldblock = 0; currsize = 0
    while currldblock < len(ldblocks):
        while currsize <= max(1,chunksize-avgldblocksize/2) and currldblock < len(ldblocks):
            currsize += ldblocks.iloc[currldblock].M_H
            currldblock += 1
        currsize = 0
        chunkendpoints += [currldblock]

    # store SNP indices of begin- and end-points of chunks
    chunkinfo = pd.DataFrame()

    # collapse data within chunks
    chunk_nums = []; chunk_denoms = []
    for n, (i,j) in enumerate(zip(chunkendpoints[:-1], chunkendpoints[1:])):
        ldblock_ind = [l for l in ldblocks.iloc[i:j].index if l in numerators.keys()]
        if len(ldblock_ind) > 0:
            chunk_nums.append(sum(
                [numerators[l] for l in ldblock_ind]))
            chunk_denoms.append(sum(
                [denominators[l] for l in ldblock_ind]))
            chunkinfo = chunkinfo.append({
                'ldblock_begin':min(ldblock_ind),
                'ldblock_end':max(ldblock_ind)+1,
                'chr_begin':ldblocks.loc[min(ldblock_ind),'chr'],
                'chr_end':ldblocks.loc[max(ldblock_ind),'chr'],
                'bp_begin':ldblocks.loc[min(ldblock_ind),'start'],
                'bp_end':ldblocks.loc[max(ldblock_ind),'end'],
                'snpind_begin':ldblocks.loc[min(ldblock_ind),'snpind_begin'],
                'snpind_end':ldblocks.loc[max(ldblock_ind),'snpind_end'],
                'numsnps':sum(ldblocks.loc[ldblock_ind,'M_H'])},
                ignore_index=True)

    ## compute leave-one-out sums
    loonumerators = []; loodenominators = []
    for i in range(len(chunk_nums)):
        loonumerators.append(sum(chunk_nums[:i]+chunk_nums[(i+1):]))
        loodenominators.append(sum(chunk_denoms[:i]+chunk_denoms[(i+1):]))

    return chunk_nums, chunk_denoms, loonumerators, loodenominators, chunkinfo

# compute estimate of effect size
def get_est(num, denom, k, num_background):
    ind = range(num_background) + [num_background+k]
    num = num[ind]
    denom = denom[ind][:,ind]
    try:
        return np.linalg.solve(denom, num)[-1]
    except np.linalg.linalg.LinAlgError:
        return np.nan

# compute standard error of estimate using jackknife
def jackknife_se(est, loonumerators, loodenominators, k, num_background):
    m = np.ones(len(loonumerators))
    theta_notj = [get_est(nu, de, k, num_background)
            for nu, de in zip(loonumerators, loodenominators)]
    g = len(m)
    n = m.sum()
    h = n/m
    theta_J = g*est - ((n-m)/n*theta_notj).sum()
    tau = est*h - (h-1)*theta_notj
    return np.sqrt(np.mean((tau - theta_J)**2/(h-1)))

# residualize the first num_background annotations out of the num_background+k-th
#   marginal annotation
# q is the numerator of the regression after background annots are residualized out
# r is the denominator of the regression after background annots are residualized out
# mux is the vector of coefficients required to residualize the background out of the
#   marginal annotatioin question
# muy is the vector of coefficients required to residualize the backgroud out of the
#   vector of GWAS summary statistics
def residualize(chunk_nums, chunk_denoms, num_background, k):
    q = np.array([num[num_background+k] for num in chunk_nums])
    r = np.array([denom[num_background+k,num_background+k] for denom in chunk_denoms])

    if num_background > 0:
        num = sum(chunk_nums)
        denom = sum(chunk_denoms)
        ATA = denom[:num_background][:,:num_background]
        ATy = num[:num_background]
        ATx = denom[:num_background][:,num_background+k]
        muy = np.linalg.solve(ATA, ATy)
        mux = np.linalg.solve(ATA, ATx)
        xiaiT = np.array([d[num_background+k,:num_background]
            for d in chunk_denoms])
        yiaiT = np.array([nu[:num_background]
            for nu in chunk_nums])
        aiaiT = np.array([d[:num_background][:,:num_background]
            for d in chunk_denoms])
        q = q - xiaiT.dot(muy) - yiaiT.dot(mux) + aiaiT.dot(muy).dot(mux)
        r = r - 2*xiaiT.dot(mux) + aiaiT.dot(mux).dot(mux)

    return q, r, mux, muy

# do sign-flipping to get p-value
def signflip(q, T, printmem=True, mode='sum'):
    def mask(a, t):
        a_ = a.copy()
        a_[np.abs(a_) < t] = 0
        return a_

    print('before sign-flipping:', fs.mem(), 'MB')

    if mode == 'sum': # use sum of q as the test statistic
        score = q.sum()
    elif mode == 'medrank': # examine how far the rank of 0 deviates from the 50th percentile
        score = np.searchsorted(np.sort(q), 0)/len(q) - 0.5
    elif mode == 'thresh': # threshold q at some absolute magnitude threshold
        top = np.percentile(np.abs(q), 75)
        print(top)
        ts = np.arange(0, top, top/10)
        q_thresh = np.array([mask(q, t) for t in ts]).T
        q_thresh /= np.linalg.norm(q_thresh, axis=0)
        scores = np.sum(q_thresh, axis=0)
        score = scores[np.argmax(np.abs(scores))]
    else:
        print('ERROR: invalid mode')
        return None

    null = np.zeros(T); current = 0; block = 100000
    while current < len(null):
        s = (-1)**np.random.binomial(1,0.5,size=(block, len(q)))
        if mode == 'sum':
            null[current:current+block] = s.dot(q)
        elif mode == 'medrank':
            null_q = s[:,:]*q[None,:]
            null_q = np.sort(null_q, axis=1)
            null[current:current+block] = \
                    np.array([np.searchsorted(s, 0)/len(s) - 0.5 for s in null_q])
        elif mode == 'thresh':
            null_q_thresh = s[:,:,None]*q_thresh[None,:,:]
            sums = np.sum(null_q_thresh, axis=1)
            null[current:current+block] = np.array([s[np.argmax(np.abs(s))] for s in sums])

        current += block
        p = max(1, ((np.abs(null) >= np.abs(score)).sum())) / float(current)
        del s; gc.collect()
        if p >= 0.01:
            null = null[:current]
            break

    se = np.abs(score)/np.sqrt(st.chi2.isf(p,1))
    del null; gc.collect()
    print('after sign-flipping:', fs.mem(), 'MB. p=', p)

    return p, score/se



import json

def add_default_params(args):
    if args.config is not None:
        # read in config file and replace '-' with '_' to match argparse behavior
        with open(args.config, 'r') as f:
            config = { k.replace('-','_') : v for k,v in json.load(f).items()}
        # overwrite with any non-None entries, resolving conflicts in favor of args
        config.update({
            k:v for k,v in args.__dict__.items()
            if (v is not None and v != []) or k not in config.keys()
            })
        # replace information in args
        args.__dict__ = config



from __future__ import print_function, division
import numpy as np

# R: SVD of (R, restricted to regression SNPs)
# R2: SVD of (R^2, restricted to regression SNPs)
def invert_weights(R, R2, sigma2g, N, x, typed=None, mode='Winv_ahat_h'):
    if typed is None:
        typed = np.isfinite(x.reshape((len(x),-1)).sum(axis=1))

    # trivial weights
    if mode == 'Winv_ahat_I':
        result = x
    # heuristic, with large-N approximation
    elif mode == 'Winv_ahat_hlN':
        U = R['U'][typed,:]; svs=R['svs']
        result = np.full(x.shape, np.nan)
        result[typed] = (U/(svs**2)).dot(U.T.dot(x[typed]))
    # heuristic, no large N approximation, using (R_o)^2 to approximate (R2)_o
    elif mode == 'Winv_ahat_h':
        U = R['U'][typed,:]; svs=R['svs']
        result = np.full(x.shape, np.nan)
        result[typed] = (U/(sigma2g*svs**2+svs/N)).dot(U.T.dot(x[typed]))
    # heuristic, no large N approximation, using R2 instead of R
    elif mode == 'Winv_ahat_h2':
        U = R2['U'][typed,:]; svs=R2['svs']
        result = np.full(x.shape, np.nan)
        result[typed] = (U/(sigma2g*svs+np.sqrt(svs)/N)).dot(U.T.dot(x[typed]))
    # exact
    elif mode == 'Winv_ahat':
        R_ = (R['U'][typed]*R['svs']).dot(R['U'][typed].T)
        R2_ = (R2['U'][typed]*R2['svs']).dot(R2['U'][typed].T)
        W = R_/N + sigma2g*R2_
        U, svs, _ = np.linalg.svd(W)
        k = np.argmax(np.cumsum(svs)/svs.sum() >= 0.95)
        # k = np.argmax(svs[:-1]/svs[1:] >= 1e5)+1
        print(R['U'].shape, R2['U'].shape, 'k=',k)
        U = U[:,:k]; svs = svs[:k]
        result = np.full(x.shape, np.nan)
        result[typed] = (U/svs).dot(U.T.dot(x[typed]))
    return result



from __future__ import print_function, division

def main(args):
    print('initializing...')
    import gzip, os, gc, time
    import numpy as np
    import pandas as pd
    import gprim.annotation as ga
    import gprim.dataset as gd
    import ypy.fs as fs
    import ypy.memo as memo
    import sldp.weights as weights

    # read in refpanel, ld blocks, and svd snps
    refpanel = gd.Dataset(args.bfile_chr)
    ldblocks = pd.read_csv(args.ld_blocks, delim_whitespace=True, header=None,
            names=['chr','start', 'end'])
    print_snps = pd.read_csv(args.print_snps, header=None, names=['SNP'])
    print_snps['printsnp'] = True
    print(len(print_snps), 'svd snps')

    # read sumstats
    print('reading sumstats', args.sumstats_stem)
    ss = pd.read_csv(args.sumstats_stem+'.sumstats.gz', sep='\t')
    ss = ss[ss.Z.notnull() & ss.N.notnull()]
    print('{} snps, {}-{} individuals (avg: {})'.format(
        len(ss), np.min(ss.N), np.max(ss.N), np.mean(ss.N)))
    ss = pd.merge(ss, print_snps[['SNP']], on='SNP', how='inner')
    print(len(ss), 'snps typed')

    # read ld scores
    print('reading in ld scores')
    ld = pd.concat([pd.read_csv(args.ldscores_chr+str(c)+'.l2.ldscore.gz',
                        delim_whitespace=True)
                    for c in range(1,23)], axis=0)
    if args.no_M_5_50:
        M = sum([int(open(args.ldscores_chr+str(c)+'.l2.M').next())
                        for c in range(1,23)])
    else:
        M = sum([int(open(args.ldscores_chr+str(c)+'.l2.M_5_50').next())
                        for c in range(1,23)])
    print(len(ld), 'snps with ld scores')
    ssld = pd.merge(ss, ld, on='SNP', how='left')
    print(len(ssld), 'hm3 snps with sumstats after merge.')

    # estimate heritability using aggregate estimator
    def esth2g(ssld):
        meanchi2 = (ssld.Z**2).mean()
        meanNl2 = (ssld.N*ssld.L2).mean()
        sigma2g = (meanchi2 - 1)/meanNl2
        h2g = sigma2g * M
        K = M/meanNl2 # h2g = K (meanchi2 - 1)
        return h2g, sigma2g, meanchi2, K
    h2g, sigma2g, meanchi2, K = esth2g(ssld)
    h2g = max(h2g, 0.03) #0.03 is an arbitrarily chosen minimum
    print('mean chi2:', meanchi2)
    print('h2g estimated at:', h2g, 'sigma2g =', sigma2g)
    if args.set_h2g:
        print('scaling Z-scores to achieve h2g of', args.set_h2g)
        norm = meanchi2 / (1 + args.set_h2g/ K)
        print('dividing all z-scores by', np.sqrt(norm))
        ssld.Z /= np.sqrt(norm)
        h2g, sigma2g, _, _ = esth2g(ssld)
        print('h2g is now', h2g)

    # write h2g results to file
    dirname = args.sumstats_stem + '.' + args.refpanel_name
    fs.makedir(dirname)
    if 1 in args.chroms:
        print('writing info file')
        info = pd.DataFrame(); info=info.append(
                {'pheno':args.sumstats_stem.split('/')[-1],
                    'h2g':h2g,
                    'sigma2g':sigma2g,
                    'Nbar':ss.N.mean()},ignore_index=True)
        info.to_csv(dirname+'/info', sep='\t', index=False)

    # preprocess ld blocks
    t0 = time.time()
    for c in args.chroms:
        print(time.time()-t0, ': loading chr', c, 'of', args.chroms)
        # get refpanel snp metadata for this chromosome
        snps = refpanel.bim_df(c)
        snps = pd.merge(snps, print_snps, on='SNP', how='left')
        snps.printsnp.fillna(False, inplace=True)
        print(len(snps), 'snps in refpanel', len(snps.columns), 'columns, including metadata')

        # merge annot and sumstats
        print('reconciling')
        snps = ga.reconciled_to(snps, ss, ['Z'],
                othercolnames=['N'], missing_val=np.nan)
        snps['typed'] = snps.Z.notnull()
        snps['ahat'] = snps.Z / np.sqrt(snps.N)

        # initialize result dataframe
        # I = no weights
        # h = heuristic weights, using R_o
        snps['Winv_ahat_I'] = np.nan # = W_o^{-1} ahat_o
        snps['R_Winv_ahat_I'] = np.nan # = R_{*o} W_o^{-1} ahat_o
        snps['Winv_ahat_h'] = np.nan # = W_o^{-1} ahat_o
        snps['R_Winv_ahat_h'] = np.nan # = R_{*o} W_o^{-1} ahat_o

        # restrict to ld blocks in this chr and process them in chunks
        for ldblock, X, meta, ind in refpanel.block_data(ldblocks, c, meta=snps):
            if meta.printsnp.sum() == 0 or \
                    not os.path.exists(args.svd_stem+str(ldblock.name)+'.R.npz'):
                print('no svd snps found in this block')
                continue
            print(meta.printsnp.sum(), 'svd snps', meta.typed.sum(), 'typed snps')
            if meta.typed.sum() == 0:
                print('no typed snps found in this block')
                snps.loc[ind, [
                    'R_Winv_ahat_I',
                    'R_Winv_ahat_h'
                    ]] = 0
                continue
            R = np.load(args.svd_stem+str(ldblock.name)+'.R.npz')
            R2 = np.load(args.svd_stem+str(ldblock.name)+'.R2.npz')
            N = meta[meta.typed.values].N.mean()
            meta_svd = meta[meta.printsnp.values]

            # multiply ahat by the weights
            x_I = snps.loc[ind[meta.printsnp],'Winv_ahat_I'] = weights.invert_weights(
                    R, R2, sigma2g, N, meta_svd.ahat.values, mode='Winv_ahat_I')
            x_h = snps.loc[ind[meta.printsnp],'Winv_ahat_h'] = weights.invert_weights(
                    R, R2, sigma2g, N, meta_svd.ahat.values, mode='Winv_ahat_h')

        print('writing processed sumstats')
        with gzip.open('{}/{}.pss.gz'.format(dirname, c), 'w') as f:
            snps.loc[snps.printsnp,['N',
                'Winv_ahat_I',
                'Winv_ahat_h'
                ]].to_csv(
                    f, index=False, sep='\t')

        del snps; memo.reset(); gc.collect()

    print('done')



from setuptools import setup

setup(name='sldp',
    version='1.1.1',
    description='Signed LD profile regression',
    url='http://github.com/yakirr/sldp',
    author='Yakir Reshef',
    author_email='yreshef@broadinstitute.org',
    license='MIT',
    packages=['sldp'],
    scripts=['sldp/sldp','sldp/preprocessannot','sldp/preprocesspheno','sldp/preprocessrefpanel'],
    install_requires=[
        'numpy',
        'pandas',
        'scipy',
        'matplotlib',
        'gprim',
        'ypy'
        ],
    zip_safe=False)
